// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

#include "manual.h"
#include <stdalign.h>
#include <stdio.h>

int main() {
    printf("%s;%zu;%zu\n", "GdkColorspace", sizeof(GdkColorspace), alignof(GdkColorspace));
    printf("%s;%zu;%zu\n", "GdkInterpType", sizeof(GdkInterpType), alignof(GdkInterpType));
    printf("%s;%zu;%zu\n", "GdkPixbufAlphaMode", sizeof(GdkPixbufAlphaMode), alignof(GdkPixbufAlphaMode));
    printf("%s;%zu;%zu\n", "GdkPixbufAnimation", sizeof(GdkPixbufAnimation), alignof(GdkPixbufAnimation));
    printf("%s;%zu;%zu\n", "GdkPixbufAnimationClass", sizeof(GdkPixbufAnimationClass), alignof(GdkPixbufAnimationClass));
    printf("%s;%zu;%zu\n", "GdkPixbufAnimationIter", sizeof(GdkPixbufAnimationIter), alignof(GdkPixbufAnimationIter));
    printf("%s;%zu;%zu\n", "GdkPixbufAnimationIterClass", sizeof(GdkPixbufAnimationIterClass), alignof(GdkPixbufAnimationIterClass));
    printf("%s;%zu;%zu\n", "GdkPixbufError", sizeof(GdkPixbufError), alignof(GdkPixbufError));
    printf("%s;%zu;%zu\n", "GdkPixbufFormat", sizeof(GdkPixbufFormat), alignof(GdkPixbufFormat));
    printf("%s;%zu;%zu\n", "GdkPixbufFormatFlags", sizeof(GdkPixbufFormatFlags), alignof(GdkPixbufFormatFlags));
    printf("%s;%zu;%zu\n", "GdkPixbufLoader", sizeof(GdkPixbufLoader), alignof(GdkPixbufLoader));
    printf("%s;%zu;%zu\n", "GdkPixbufLoaderClass", sizeof(GdkPixbufLoaderClass), alignof(GdkPixbufLoaderClass));
    printf("%s;%zu;%zu\n", "GdkPixbufModule", sizeof(GdkPixbufModule), alignof(GdkPixbufModule));
    printf("%s;%zu;%zu\n", "GdkPixbufModulePattern", sizeof(GdkPixbufModulePattern), alignof(GdkPixbufModulePattern));
    printf("%s;%zu;%zu\n", "GdkPixbufRotation", sizeof(GdkPixbufRotation), alignof(GdkPixbufRotation));
    return 0;
}
